
/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#undef FILENAME
#define FILENAME "initng_parser.c"
#include "initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>

/*#define DEBUG*/
                                                                                                                                                                                                                                                                                                                                                                        /*#include <dirent.h> *//* opendir() closedir() */

#include "string_tools.h"
#include "service_db.h"
#include "initng_parser.h"
#include "open_read_close.h"

static service_h *parse_file(char *filename, char *watch_for);
static int set_parser(e_sdt type, char **value, service_h * from_service);
static int disable_parser(e_sdt type, char **value, service_h * from_service);
static int up_when_parser(e_sdt type, char **value, service_h * from_service);
static int string_parser(e_sdt type, char **value, service_h * from_service);
static int strings_parser(e_sdt type, char **value, service_h * from_service);
static int script_parser(e_sdt type, char **value, service_h * from_service);

/*static int is_true(char *value);*/
static int is_false(char *value);

typedef struct
{
    e_sdt type;
    const char *opt_name;
    int (*parse_func) (e_sdt t, char **where, service_h * f);
} option_row;

static const option_row option_table[] = {
    {STRING_USE, "after", strings_parser},
    {STRING_NEED_ALSO, "also_start", strings_parser},
    {STRING_DAEMON_ARGS, "daemon_args", string_parser},
    {STRING_DAEMON, "daemon", string_parser},
    {STRING_NEED, "depends", strings_parser},
    {STRING_NEED, "depend", strings_parser},
    {STRING_NEED, "needs", strings_parser},
    {STRING_NEED, "need", strings_parser},
    {ENABLED_RESPAWN, "norespawn", disable_parser},
    {STRING_USE, "uses", strings_parser},
    {STRING_PROVIDE, "provides", strings_parser},
    {STRING_PROVIDE, "provide", strings_parser},
    {STRING_PID_FILE, "pid_file", string_parser},
    {ENABLED_RESPAWN, "respawn", set_parser},
    {STRING_RDEPENDS, "rdepends", strings_parser},
    {STRING_RDEPENDS, "rdepend", strings_parser},
    {STRING_START_SCRIPT, "start\n{", script_parser},
    {STRING_START_SCRIPT, "start {", script_parser},
    {STRING_START_SCRIPT, "start{", script_parser},
    {STRING_START_ARGS, "start_args", string_parser},
    {STRING_START, "start", string_parser},
    {STRING_STOP_SCRIPT, "stop\n{", script_parser},
    {STRING_STOP_SCRIPT, "stop {", script_parser},
    {STRING_STOP_SCRIPT, "stop{", script_parser},
    {STRING_STOP_ARGS, "stop_args", string_parser},
    {STRING_STOP, "stop", string_parser},
    {EMPTY, "up_when", up_when_parser},
    {STRING_USE, "use", strings_parser},
    {EMPTY, "when_up", up_when_parser},
    {STRING_STDOUT, "stdout", string_parser},
    {STRING_STDOUT, "std_out", string_parser},
    {STRING_STDERR, "stderr", string_parser},
    {STRING_STDERR, "std_err", string_parser},
    {STRING_ENV, "environ", strings_parser},
    {STRING_ENV, "env", strings_parser},

    {EMPTY, NULL, NULL}
};

static int set_parser(e_sdt type, char **value, service_h * from_service)
{
    int line_lenght;

    if (!value)
    {
        printf("set_parser(), did not get value ponter.\n");
        return (FALSE);
    }
    if (!from_service)
    {
        printf("set_parser(): did not get a service to set data into.\n");
        return (FALSE);
    }
#ifdef DEBUG
    printf("enable_parser(%i,s,%s);\n", type, from_service->name);
#endif
    line_lenght = strcspn(*value, "\n;");       /* Get line lenght */
    if (line_lenght && is_false(*value))
    {
        service_db_remove(type, from_service);
    }
    else
    {
        /* else set true */
        service_db_set(type, from_service);
    }
    return (TRUE);
}

static int disable_parser(e_sdt type, char **value, service_h * from_service)
{
    int line_lenght;

    assert(from_service);
#ifdef DEBUG
    printf("disable_parser(%i,%s,%s);\n", type, *value, from_service->name);
#endif
    line_lenght = strcspn(*value, "\n;");       /* Get line lenght */
    if (!(line_lenght && is_false(*value)))
    {
        /* else set true */
        service_db_set(type, from_service);
    }
    return (TRUE);

}

static int up_when_parser(e_sdt type, char **value, service_h * from_service)
{
    assert(from_service);
#ifdef DEBUG
    printf("up_when_parser(%i,s,%s);\n", type, from_service->name);
#endif
    if (!*value && type == EMPTY)
        return (FALSE);
    st_jump_spaces(value);
    /* only one can be enabled, disable all others */
    service_db_remove(ENABLED_UP_WHEN_DONE, from_service);
    service_db_remove(ENABLED_UP_WHEN_PID_SET, from_service);
    service_db_remove(ENABLED_UP_WHEN_STARTED, from_service);

    if (strncasecmp(*value, "started", 7) == 0 ||
        strncasecmp(*value, "start", 5) == 0)
        service_db_set(ENABLED_UP_WHEN_STARTED, from_service);

    else if (strncasecmp(*value, "done", 4) == 0 ||
             strncasecmp(*value, "exit", 4) == 0)
        service_db_set(ENABLED_UP_WHEN_DONE, from_service);

    else if (strncasecmp(*value, "pid_set", 7) == 0 ||
             strncasecmp(*value, "pidset", 6) == 0 ||
             strncasecmp(*value, "pid_is_set", 10) == 0)
        service_db_set(ENABLED_UP_WHEN_PID_SET, from_service);

    else
    {
        printf("Fail parsing up_when (%s)\n", from_service->name);
        return (FALSE);
    }
    return (TRUE);
}

static int string_parser(e_sdt type, char **value, service_h * from_service)
{
    char *to;

    assert(from_service);
    if (!(*value))
        return (FALSE);
#ifdef DEBUG
    printf("string_parser(%i,s,%s);\n", type, from_service->name);
#endif
    if ((to = st_dup_line(value)))
    {
        service_db_set_string(type, from_service, to);
        return (TRUE);
    }
    return (FALSE);
}

static int strings_parser(e_sdt type, char **value, service_h * from_service)
{
    char *to = NULL;

    assert(from_service);
    if (!(*value))
        return (FALSE);
#ifdef DEBUG
    printf("string_parser(%i,s,%s);\n", type, from_service->name);
#endif

    while ((to = st_dup_next_word(value)))
    {
        service_db_set_another_string(type, from_service, to);
        st_jump_spaces(value);
    }
    return (TRUE);
    /*return(FALSE); */
}

static int script_parser(e_sdt type, char **value, service_h * from_service)
{
    int stack_lenght = 0;

    assert(from_service);
#ifdef DEBUG
    printf("script_parser(%i,s,%s);\n", type, from_service->name);
#endif
    stack_lenght = chars_to_end_of_this_stack(*value);

    service_db_set_string(type, from_service, strndup(*value, stack_lenght));
    (*value) = (*value) + stack_lenght;
    return (TRUE);
}

static int parse_opt(char **where, service_h * srv)
{
    int i;
    int option_word_lenght = 0;

    /* jump forward all spaces */
    st_jump_spaces(where);

    for (i = 0; option_table[i].opt_name != NULL; i++)
    {
        if (strncasecmp
            (option_table[i].opt_name, *where,
             (option_word_lenght = strlen(option_table[i].opt_name))) == 0)
        {
#ifdef DEBUG
            printf("parse_opt(%s): option \"%s\" value at option no %i.\n",
                   srv->name, option_table[i].opt_name, i);
#endif
            *where = *where + option_word_lenght;
            /* jump forward all spaces */
            st_jump_spaces(where);
            return (option_table[i].
                    parse_func(option_table[i].type, where, srv));
        }

    }

    /* Go here if there is errors! */
    {
        char *line = strndup(*where, strcspn(*where, "\n\0"));

        printf("parse_opt(%s): line: \"%s\"!\n", srv->name, line);
        free(line);
    }
    printf("sleeping 2 seconds......\n");
    sleep(2);
    return (FALSE);
}

/*
static int is_true(char *value) {
    while(value[0]==' ' || value[0]=='=') value++;
    if(strncasecmp(value,"on",2)==0 ||
       strncasecmp(value,"yes",3)==0 ||
       strncasecmp(value,"true",4)==0)
        return(TRUE);
    return(FALSE);
}
*/

static int is_false(char *value)
{
    while (value[0] == ' ' || value[0] == '=')
        value++;
    if (strncasecmp(value, "no", 2) == 0 ||
        strncasecmp(value, "off", 3) == 0 ||
        strncasecmp(value, "false", 5) == 0)
        return (TRUE);
    return (FALSE);
}

static service_h *parse_service_line(char **to_parse, char *watch_for,
                                     service_h * father, e_type type)
{
    char *name = NULL;
    char *copy = NULL;
    service_h *new_service = NULL;
    service_h *service_to_return = NULL;

    assert(to_parse);
    assert(*to_parse);
    assert(watch_for);

#ifdef DEBUG
    printf("parse_service_line(*s*,%s);\n", watch_for);
#endif

    /* service test : class { */
    /*        |               */
    /* jump forward on spaces */
    st_jump_spaces(to_parse);

    /* service test : class { */
    /*         |              */

    /* fetch name */
    if (!(name = st_dup_next_word(to_parse)))
    {
        printf("Did not got name!\n");
        return (NULL);
    }
    /* service test : class { */
    /*             |          */

#ifdef DEBUG
    printf("parse_service_line(s,%s): service name: \"%s\"\n", watch_for,
           name);
#endif

    /* jump forward on spaces */
    st_jump_spaces(to_parse);

    /* service test : class { */
    /*              |         */
    if ((*to_parse)[0] == ':')
    {
        (*to_parse)++;
        /* service test : class { */
        /*               |        */

        /* jump forward on spaces */
        st_jump_spaces(to_parse);
        /* service test : class { */
        /*                |       */

        copy = st_dup_next_word(to_parse);
        father = service_db_find_by_name(copy);
        if (!father)
        {
            printf("father %s not found!\n", copy);
            free(copy);
            return (NULL);
        }
        printf("father is: %s.\n", father->name);
        free(copy);

        /* service test : class { */
        /*                     |  */
        /* jump forward on spaces */
        st_jump_spaces(to_parse);

    }

    /* service test : class { */
    /*                      | */
    if ((*to_parse)[0] != '{')
    {
        /* something is wrong */
        printf("error parsing, expected something else.\n");
        return (NULL);
    }

    /* jump to first char after start tag and begin from there. */
    (*to_parse)++;
    /* service test : class { */
    /*                       | */

    /* create a service entry */
    if (father)
        service_db_copy(&new_service, name, father);
    else
        service_db_new(&new_service, name, type);
    if (!new_service)
    {
        printf("parse_service_line(s,s): unable to allocate new service!\n");
        free(name);
        return (NULL);
    }

    /* set type if not set. */
    if (father)
        new_service->type = type;

    /* skip lines with '#' */
    st_jump_to_valid_line(to_parse);

    /* cary on until segment stop or eof */
    while (*to_parse[0] != '\0' && *to_parse[0] != '}')
    {
        service_h *tmp = NULL;

        /* if this row has keyword service */
        if (st_cmp(to_parse, "service"))
            if ((tmp =
                 parse_service_line(to_parse, watch_for, NULL, TYPE_SERVICE)))
                service_to_return = tmp;

        /* if this row has keyword class */
        if (st_cmp(to_parse, "class"))
            if ((tmp =
                 parse_service_line(to_parse, watch_for, NULL, TYPE_CLASS)))
                service_to_return = tmp;

        /* parse line - search for keywords in option_table */
        parse_opt(to_parse, new_service);

        /* jump to next row */
        st_jump_to_next_row(to_parse);
    }

    /* add to service db */
    if (!service_db_add(new_service))
    {
        printf("can't add service to service db!\n");
        service_db_free(new_service);
        return (NULL);
    }

    /* check if this is the service we ar looking for */
    if (st_service_cmp(watch_for, new_service->name))
        return (new_service);

    /* if we got a service from a under service */
    if (service_to_return)
        return (service_to_return);
    /* return null if not */
    return (NULL);
}

/* parse a file for services */
static service_h *parse_file(char *filename, char *watch_for)
{
    char *file, *orig;
    service_h *service_to_return = NULL;
    service_h *tmp = NULL;

    assert(filename);
    assert(watch_for);

#ifdef DEBUG
    printf("parse_file(%s);\n", filename);
#endif

    /* read config file */
    if (!open_read_close(filename, &orig))
    {
#ifdef DEBUG
        printf("Can't open config file %s.\n", filename);
#endif
        return (NULL);
    }
    file = orig;

    /* skip lines starting with '#' */
    st_jump_to_valid_line(&file);

    /* main parse line by line loop */
    while (file[0] != '\0')
    {

        /* if this row has keyword service */
        if (st_cmp(&file, "service"))
            if ((tmp =
                 parse_service_line(&file, watch_for, NULL, TYPE_SERVICE)))
                service_to_return = tmp;

        /* if this row has keyword class */
        if (st_cmp(&file, "class"))
            if ((tmp =
                 parse_service_line(&file, watch_for, NULL, TYPE_CLASS)))
                service_to_return = tmp;

        /* jump file pointer to first char next row */
        st_jump_to_next_row(&file);
    }                           /* end while */

    free(orig);
#ifdef DEBUG_PRINT
    print_all();
#endif
    /*exit_initng(1); */
    if (service_to_return)
        return (service_to_return);
    return (NULL);
}

/* Load a service from a process_name or process_path */
service_h *initng_load_process(char *process_name)
{
    char filetoparse[200];
    char *tmp = NULL;
    service_h *got_serv = NULL;

    assert(process_name);

    /* example  INITNG_ROOT=/etc/initng process_name="deamon/sshd INITNG_EXT=".i" */
    /* TRY NO 1, try load   "/etc/initng/daemon/sshd.i" */
    /* TRY NO 2, try load   "/etc/initng/daemon/sshd/sshd.i" */
    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.i" */
    /* TRY NO 4, try load   "/etc/initng/daemon/default.i" */
    /* TRY NO 5, try load   "/etc/initng/daemon/daemon.i */
    /* TRY NO 6, try load   "/etc/initng/sshd.i" */
    /* TRY NO 7, try load   "/etc/initng/daemon.i */

    /* TRY NO 1, try load   "/etc/initng/daemon/sshd/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, "/");
    strcat(filetoparse, st_strip_path(process_name));
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 2, try load   "/etc/initng/daemon/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, "/default" INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 4, try "/etc/initng/daemon/default.i" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, "/default" INITNG_EXT);
    free(tmp);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 5, try "/etc/initng/daemon/daemon.i" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, INITNG_EXT);
    free(tmp);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 6, try "/etc/initng/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, st_strip_path(process_name));
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 7, try "/etc/initng/daemon.i" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, INITNG_EXT);
    free(tmp);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

#ifdef DEBUG
    printf(" ** FAIL " FILENAME " initng_load_process(%s): got \"NULL\"\n",
           process_name);
#endif
    return (NULL);
}
